# File kendoku.py
from backtrack import Backtrack
import sys
import functools
from functools import reduce

class Kendoku(Backtrack):
    # Beispiel: https://www.janko.at/Raetsel/Kendoku/009.a.htm
    

    _kendoku = {} # Interne Darstellung als dict
    
    # Zählen der Lösungen
    solutionNr = 0  

    # Typbeschreibungen
    Places = tuple[int]
    Kendokutuple = tuple[int, str, Places]
    Moves = list[int]

    def __init__(self, size=0, kendoku=(), test:bool= False, showAttempts:bool = False):
        if size<=0:
            print(f"size ist ungültig: {size}")
            sys.exit()
        self.size=size
        self.convert(kendoku)

        # Zwischenergebnisse und Lösung
        # Obwohl eine Zahlenmatrix gesucht wird, wird zur internen Berechnung
        # ein eindimensionaler Vektor verwerdet.
        # Zeilen- und Spaltennummern werden mit der Funktion rowAndCol
        # aus der Platznummer (level) errechnet

        self.vector = [None for i in range(size*size)]

        # Welche Werte können in den einzelnen Feldern stehen?
        self.values = set(range(1, self.size+1))   

        # In rset und cset werden jene Werte gespeichert, die in der jeweiligen Zeile
        # oder Spalte schon vorkommen und daher in moves nicht mehr zur Auswahl stehen
        self.rset = [set() for i in range(size)]  # row sets
        self.cset = [set() for i in range(size)]  # column sets
       
        super().__init__(test, showAttempts)        

    # Für die Prüfung eines Bereichs wird kendoku in eine
    # interne Darstellung (_kendoku) als dict mit dem höchsten Index
    # als key umgewandelt    
    def convert(self, kendoku:Kendokutuple):
        # suche die höchste Feldnummer und verwende sie als Index
        for k in kendoku:
            # Höchste Platznummer als key festlegen
            key = max(k[2])

            # Durch die Verwendung eines dict lesbarer machen
            self._kendoku[key] = {"value":k[0], "op":k[1], "members":k[2]}

    # Die Hilfsfunktion roeAndCol errechnet aus der Platznummer (where)
    # die Zeilennummer (r) und die Spaltennummer (c)
    def rowAndCol(self, where: int):
        s = self.size
        r = where // s
        c = where % s
        return (r, c)

    # Bestimme die Werte, die für den Platz (level) zulässig sind    
    def moves(self, level: int) -> Moves:        
        r, c = self.rowAndCol(level)
        # Werte, die in derselben Spalte oder Zeile vorkommen, werden ausgeschlossen
        return list(self.values - self.rset[r] - self.cset[c])

    # Ist der Wert (move) auf dem vorgesehenen Platz (level) erlaubt?
    def valid(self, level: int, move: int) -> bool:
        
        # Falls der Platz nicht der letzte des Bereichs ist: weitersuchen!
        # Dazu prüft get, ob level ein gültiger Key von _kendoku ist.
        
        k = self._kendoku.get(level, False)

        # Wenn nein, kann die Rechenopration in dem 
        # Bereich noch nicht ausgeführt werden. Weitersuchen!

        if not k:  
            return True

        # Ist der key vorhanden, dann kann der Test beginnen
        # Die Nummern der Felder des Bereichs sind in members zu finden.
        # Nun müssen zur Berechnung alle Werte des Bereichs
        # (ohne das letzte) in der Liste area zwischengespeichert werden.
        
        area = []
        for i in k["members"][0:-1]:  
            area.append(self.vector[i])
            
        # Das letzte Feld wird probeweise dazu genommen
        area.append(move)

        # Nun wird die Berechnung in dem Bereich ausgeführt.
        # Berechnungen, abhängig vom Operator (+ - ...)
        value = k["value"]
        match(k["op"]):        
            case "*" | "x" | "X":
                return reduce(lambda x,y:x*y, area)==value
            case "/" | ":":
                return area[0]//area[1]==value or\
                    area[1]//area[0]==value
            case "+":
                return reduce(lambda x,y:x+y,area)==value
            case "-":
                return abs(area[0]-area[1])==value
            case _:  # Keine Operation angegeben
                return True
        
    
    # Überprüften (validierten) Wert (move) auf den Platz (level) speichern
    def record(self, level: int, move: int):        
        self.vector[level]=move
        r, c = self.rowAndCol(level)
        self.rset[r].add(move)  # Wert zur Menge der Zeilenwerte hinzufügen
        self.cset[c].add(move)  # Wert zur Menge der Spaltenwerte hinzufügen
       
    # Der Wert (move) passt nicht, daher vom Platz (level) entfernen
    def undo(self, level: int, move: int):       
        self.vector[level]=None     
        r, c = self.rowAndCol(level)
        self.rset[r].remove(move)  # Wert aus der Menge der Zeilenwerte entfernen
        self.cset[c].remove(move)  # Wert aus der Menge der Spaltenwerte entfernen        

    # Wurde eine Lösung gefunden?
    def done(self, level: int):
        if level==self.size*self.size-1:
            self.solutionNr += 1
            print(f"Lösung {self.solutionNr:4}")
            for r in range(self.size):
                for c in range(self.size):
                    print(f"{self.vector[r*self.size + c]:4} ", end="")
                print()
            return True
        return False

if __name__ == "__main__":
    """
    Jeder Kendoku-Bereich wird durch ein Tupel codiert:
    - Der erwartete Wert
    - Die Rechenoperation (zulässig: + - x X * / )
    - Die Nummern der Felder, die zu dem Bereich gehören, als Tupel.
      Jede Feldnummer muss in genau einem der Tupel genau einmal vorkommen
    """
    kendoku = (
        (16, "x", ( 0,  1,  6)),
        ( 7, "+", ( 2,  3,  9)),
        ( 1, "-", ( 4,  5)),
        (25, "x", ( 7,  8, 13)),
        ( 2, "/", (10, 11)),
        ( 2, "/", (12, 18)),
        ( 2, "-", (14, 20)),
        ( 4, "-", (15, 21)),
        ( 6, "x", (16, 17)),
        ( 4, "-", (19, 25)),
        ( 2, "/", (22, 28)),
        ( 5, "+", (23, 29)),
        ( 3, "-", (24, 30)),
        ( 9, "x", (26, 31, 32)),
        ( 2, "-", (27, 33)),
        ( 1, "-", (34, 35))
    )
    
    Kendoku(size=6, kendoku=kendoku, showAttempts=True)
